<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace MapModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Table\HostgroupsTable;
use App\Model\Table\HoststatusTable;
use App\Model\Table\ServicegroupsTable;
use App\Model\Table\ServicestatusTable;
use Cake\Datasource\EntityInterface;
use Cake\ORM\Association\HasMany;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\Query;
use Cake\ORM\Table;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Core\Hoststatus;
use itnovum\openITCOCKPIT\Core\Servicestatus;
use itnovum\openITCOCKPIT\Core\Views\BBCodeParser;
use itnovum\openITCOCKPIT\Core\Views\ServiceStateSummary;
use itnovum\openITCOCKPIT\Core\Views\UserTime;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\MapFilter;
use MapModule\Model\Entity\Map;
use Statusengine\PerfdataParser;

/**
 * Maps Model
 *
 * @property MapgadgetsTable&HasMany $Mapgadgets
 * @property MapiconsTable&HasMany $Mapicons
 * @property MapitemsTable&HasMany $Mapitems
 * @property MaplinesTable&HasMany $Maplines
 * @property RotationsTable&HasMany $MapsToRotations
 * @property MapsummaryitemsTable&HasMany $Mapsummaryitems
 * @property MaptextsTable&HasMany $Maptexts
 *
 * @method Map get($primaryKey, $options = [])
 * @method Map newEntity($data = null, array $options = [])
 * @method Map[] newEntities(array $data, array $options = [])
 * @method Map|false save(EntityInterface $entity, $options = [])
 * @method Map saveOrFail(EntityInterface $entity, $options = [])
 * @method Map patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method Map[] patchEntities($entities, array $data, array $options = [])
 * @method Map findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin TimestampBehavior
 */
class MapsTable extends Table {

    use PaginationAndScrollIndexTrait;

    private $hostIcons = [
        0 => 'up.png',
        1 => 'down.png',
        2 => 'unreachable.png'
    ];
    private $serviceIcons = [
        0 => 'up.png',
        1 => 'warning.png',
        2 => 'critical.png',
        3 => 'unknown.png'
    ];
    private $ackIcon = 'ack.png';
    private $downtimeIcon = 'downtime.png';
    private $ackAndDowntimeIcon = 'downtime_ack.png';

    private $errorIcon = 'error.png';

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('maps');
        $this->setDisplayField('name');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->hasMany('Mapgadgets', [
            'foreignKey' => 'map_id',
            'className'  => 'MapModule.Mapgadgets'
        ])->setDependent(true);

        $this->hasMany('Mapicons', [
            'foreignKey' => 'map_id',
            'className'  => 'MapModule.Mapicons'
        ])->setDependent(true);

        $this->hasMany('Mapitems', [
            'foreignKey' => 'map_id',
            'className'  => 'MapModule.Mapitems'
        ])->setDependent(true);

        $this->hasMany('Maplines', [
            'foreignKey' => 'map_id',
            'className'  => 'MapModule.Maplines'
        ])->setDependent(true);

        $this->hasMany('MapsToRotations', [
            'foreignKey' => 'map_id',
            'className'  => 'MapModule.MapsToRotations'
        ])->setDependent(true);

        $this->hasMany('Mapsummaryitems', [
            'foreignKey' => 'map_id',
            'className'  => 'MapModule.Mapsummaryitems'
        ])->setDependent(true);

        $this->hasMany('Maptexts', [
            'foreignKey' => 'map_id',
            'className'  => 'MapModule.Maptexts'
        ])->setDependent(true);

    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        return $validator;
    }

    /**
     * @param MapFilter $MapFilter
     * @param null|PaginateOMat $PaginateOMat
     * @return array
     */
    public function getMapsIndex(MapFilter $MapFilter, $PaginateOMat = null) {
        $query = $this->find('all')
            ->where($MapFilter->indexFilter())
            ->distinct('Maps.id')
            ->order($MapFilter->getOrderForPaginator('Maps.name', 'asc'));

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['Maps.id' => $id]);
    }

    /**
     * @param $realMapId
     * @param $mapItemMapId
     * @return array
     */
    public function getMapsForMaps($realMapId, $mapItemMapId) {
        $query = $this->find()
            ->join([
                'table'      => 'mapitems',
                'type'       => 'INNER',
                'alias'      => 'Mapitems',
                'conditions' => 'Mapitems.object_id = Maps.id',
            ])
            ->where([
                'Maps.id'         => $realMapId,
                'Mapitems.map_id' => $mapItemMapId,
            ]);

        $result = $query->first();
        if (empty($result)) {
            return [];
        }
        return $query->toArray();
    }

    /**
     * @param $realMapId
     * @param $mapItemMapId
     * @return array
     */
    public function getMapsForMapsummaryitems($realMapId, $mapItemMapId) {
        $query = $this->find()
            ->join([
                'table'      => 'mapsummaryitems',
                'type'       => 'INNER',
                'alias'      => 'Mapsummaryitems',
                'conditions' => 'Mapsummaryitems.object_id = Maps.id',
            ])
            ->where([
                'Maps.id'                => $realMapId,
                'Mapsummaryitems.map_id' => $mapItemMapId,
            ]);

        $result = $query->first();
        if (empty($result)) {
            return [];
        }
        return $query->toArray();
    }

    /**
     * @param $object_id
     * @return array
     */
    public function getMapsummaryitemForMapsummary($object_id) {
        $query = $this->find()
            ->join([
                'table'      => 'mapsummaryitems',
                'type'       => 'INNER',
                'alias'      => 'Mapsummaryitems',
                'conditions' => 'Mapsummaryitems.object_id = Maps.id',
            ])
            ->select([
                'Mapsummaryitems.object_id',
                'Mapsummaryitems.map_id',
            ])
            ->where([
                'Mapsummaryitems.object_id' => $object_id,
            ])->enableAutoFields(true);

        $result = $query->first();
        if (empty($result)) {
            return [];
        }
        return $query->toArray();
    }

    /**
     * @param $object_id
     * @return array
     */
    public function getMapitemForMapsummary($object_id) {
        $query = $this->find()
            ->join([
                'table'      => 'mapitems',
                'type'       => 'INNER',
                'alias'      => 'Mapitems',
                'conditions' => 'Mapitems.object_id = Maps.id',
            ])
            ->select([
                'Mapitems.object_id',
                'Mapitems.map_id',
            ])
            ->where([
                'Mapitems.object_id' => $object_id,
            ])->enableAutoFields(true);

        $result = $query->first();
        if (empty($result)) {
            return [];
        }
        return $query->toArray();
    }


    /**
     * @param ServicestatusTable $ServicestatusTable
     * @param \App\Model\Entity\Hoststatus $hoststatus
     * @return array|void
     */
    public function getHostInformation(ServicestatusTable $ServicestatusTable, \App\Model\Entity\Hoststatus $hoststatus) {
        $HoststatusView = new Hoststatus($hoststatus);

        if (empty($hoststatus)) {
            $HoststatusView = new Hoststatus([]);
            return [
                'icon'           => $this->errorIcon,
                'icon_property'  => $this->errorIcon,
                'isAcknowledged' => false,
                'isInDowntime'   => false,
                'color'          => 'text-primary',
                'background'     => 'bg-not-monitored',
                'Hoststatus'     => $HoststatusView->toArray(),
                'Host'           => $hoststatus->get('host_mapping')->toArray()
            ];
        }


        $icon = $this->hostIcons[$HoststatusView->currentState()];
        $color = $HoststatusView->HostStatusColor();
        $background = $HoststatusView->HostStatusBackgroundColor();

        $iconProperty = $icon;
        if ($HoststatusView->isAcknowledged()) {
            $iconProperty = $this->ackIcon;
        }

        if ($HoststatusView->isInDowntime()) {
            $iconProperty = $this->downtimeIcon;
        }

        if ($HoststatusView->isAcknowledged() && $HoststatusView->isInDowntime()) {
            $iconProperty = $this->ackAndDowntimeIcon;
        }

        $return = [
            'icon'           => $icon,
            'icon_property'  => $iconProperty,
            'isAcknowledged' => $HoststatusView->isAcknowledged(),
            'isInDowntime'   => $HoststatusView->isInDowntime(),
            'color'          => $color,
            'background'     => $background,
            'Hoststatus'     => $HoststatusView->toArray(),
            'Host'           => $hoststatus->get('host_mapping')->toArray()
        ];


        if ($HoststatusView->currentState() > 0) {
            return $return;
        }

        //Check services for cumulated state (only if host is up)
        $servicesStatus = $ServicestatusTable->getServicesByHostUuid($hoststatus->get('hostname'));

        if (empty($servicesStatus)) {
            return $return;
        }


        $worstServiceState = array_values(
            Hash::sort($servicesStatus, '{n}.current_state', 'desc')
        );

        $servicestatus = new Servicestatus($worstServiceState[0]);
        $serviceIcon = $this->serviceIcons[$servicestatus->currentState()];

        $serviceIconProperty = $serviceIcon;
        if ($servicestatus->isAcknowledged()) {
            $serviceIconProperty = $this->ackIcon;
        }

        if ($servicestatus->isInDowntime()) {
            $serviceIconProperty = $this->downtimeIcon;
        }

        if ($servicestatus->isAcknowledged() && $servicestatus->isInDowntime()) {
            $serviceIconProperty = $this->ackAndDowntimeIcon;
        }

        return [
            'icon'           => $serviceIcon,
            'icon_property'  => $serviceIconProperty,
            'isAcknowledged' => $servicestatus->isAcknowledged(),
            'isInDowntime'   => $servicestatus->isInDowntime(),
            'color'          => $servicestatus->ServiceStatusColor(),
            'background'     => $servicestatus->ServiceStatusBackgroundColor(),
            'Hoststatus'     => $HoststatusView->toArray(),
            'Host'           => $hoststatus->get('host_mapping')->toArray()
        ];
    }

    /**
     * @param \App\Model\Entity\Servicestatus $servicestatus
     * @param \App\Model\Entity\Hoststatus $hoststatus
     * @param bool $includeServiceOutput
     * @return array
     */
    public function getServiceInformation(\App\Model\Entity\Servicestatus $servicestatus, \App\Model\Entity\Hoststatus $hoststatus, $includeServiceOutput = false) {
        if (empty($servicestatus)) {
            $ServicestatusView = new Servicestatus([]);
            return [
                'icon'           => $this->errorIcon,
                'icon_property'  => $this->errorIcon,
                'isAcknowledged' => false,
                'isInDowntime'   => false,
                'color'          => 'text-primary',
                'background'     => 'bg-not-monitored',
                'Servicestatus'  => $ServicestatusView->toArray(),
                'Perfdata'       => [],
                'Host'           => $hoststatus->get('host_mapping')->toArray(),
                'Service'        => $servicestatus->get('service_mapping')->toArray()
            ];
        }
        $ServicestatusView = new Servicestatus($servicestatus);

        $icon = $this->serviceIcons[$ServicestatusView->currentState()];

        $iconProperty = $icon;
        if ($ServicestatusView->isAcknowledged()) {
            $iconProperty = $this->ackIcon;
        }

        if ($ServicestatusView->isInDowntime()) {
            $iconProperty = $this->downtimeIcon;
        }

        if ($ServicestatusView->isAcknowledged() && $ServicestatusView->isInDowntime()) {
            $iconProperty = $this->ackAndDowntimeIcon;
        }

        $perfdata = new PerfdataParser($ServicestatusView->getPerfdata());

        $tmpServicestatus = $ServicestatusView->toArray();
        if ($includeServiceOutput === true) {
            $Parser = new BBCodeParser();
            $tmpServicestatus['output'] = h($ServicestatusView->getOutput());
            $tmpServicestatus['longOutputHtml'] = $Parser->nagiosNl2br($Parser->asHtml($ServicestatusView->getLongOutput(), true));
        }

        return [
            'icon'           => $icon,
            'icon_property'  => $iconProperty,
            'isAcknowledged' => $ServicestatusView->isAcknowledged(),
            'isInDowntime'   => $ServicestatusView->isInDowntime(),
            'color'          => $ServicestatusView->ServiceStatusColor(),
            'background'     => $ServicestatusView->ServiceStatusBackgroundColor(),
            'Perfdata'       => $perfdata->parse(),
            'Servicestatus'  => $tmpServicestatus,
            'Host'           => $hoststatus->get('host_mapping')->toArray(),
            'Service'        => $servicestatus->get('service_mapping')->toArray()
        ];
    }

    /**
     * @param array $hostgroup
     * @param HoststatusTable $HoststatusTable
     * @param ServicestatusTable $ServicestatusTable
     * @return array
     */
    public function getHostgroupInformation(array $hostgroup, HoststatusTable $HoststatusTable, ServicestatusTable $ServicestatusTable) {
        $hostUuids = Hash::extract($hostgroup['hostgroup_members'], '{n}.host_uuid');

        $hoststatusByUuids = $HoststatusTable->byUuids($hostUuids, false);
        $hostgroupLight = [
            'uuid'        => $hostgroup['uuid'],
            'name'        => $hostgroup['HostgroupMapping']['display_name'],
            'description' => $hostgroup['description']
        ];

        if (empty($hoststatusByUuids)) {
            return [
                'icon'       => $this->errorIcon,
                'color'      => 'text-primary',
                'background' => 'bg-not-monitored',
                'Hostgroup'  => $hostgroupLight
            ];
        }
        $worstHostState = array_values(
            Hash::sort($hoststatusByUuids, '{s}.current_state', 'desc')
        );


        $hoststatus = new Hoststatus($worstHostState[0]);

        $icon = $this->hostIcons[$hoststatus->currentState()];
        $color = $hoststatus->HostStatusColor();
        $background = $hoststatus->HostStatusBackgroundColor();


        if ($hoststatus->isAcknowledged()) {
            $icon = $this->ackIcon;
        }

        if ($hoststatus->isInDowntime()) {
            $icon = $this->downtimeIcon;
        }

        if ($hoststatus->isAcknowledged() && $hoststatus->isInDowntime()) {
            $icon = $this->ackAndDowntimeIcon;
        }

        if ($hoststatus->currentState() > 0) {
            return [
                'icon'       => $icon,
                'color'      => $color,
                'background' => $background,
                'Hostgroup'  => $hostgroupLight
            ];
        }


        //Check services for cumulated state (only if host is up)
        $services = $ServicestatusTable->byHostUuids($hostUuids, false, ['Servicestatus.current_state IN' => [1, 2, 3]]);
        if (!empty($services)) {
            $worstServiceState = array_values(
                Hash::sort($services, '{s}.current_state', 'desc')
            );

            $servicestatus = new Servicestatus($worstServiceState[0]);
            $serviceIcon = $this->serviceIcons[$servicestatus->currentState()];

            if ($servicestatus->isAcknowledged()) {
                $serviceIcon = $this->ackIcon;
            }

            if ($servicestatus->isInDowntime()) {
                $serviceIcon = $this->downtimeIcon;
            }

            if ($servicestatus->isAcknowledged() && $servicestatus->isInDowntime()) {
                $serviceIcon = $this->ackAndDowntimeIcon;
            }
            return [
                'icon'       => $serviceIcon,
                'color'      => $servicestatus->ServiceStatusColor(),
                'background' => $servicestatus->ServiceStatusBackgroundColor(),
                'Hostgroup'  => $hostgroupLight
            ];
        }

        return [
            'icon'       => $icon,
            'color'      => $color,
            'background' => $background,
            'Hostgroup'  => $hostgroupLight
        ];
    }

    /**
     * @param ServicestatusTable $ServicestatusTable
     * @param array $servicegroup
     * @return array
     */
    public function getServicegroupInformation(ServicestatusTable $ServicestatusTable, array $servicegroup) {
        $serviceUuids = Hash::extract($servicegroup['servicegroup_members'], '{n}.service_uuid');

        $servicestatusByUuids = [];
        if (!empty($serviceUuids)) {
            $servicestatusByUuids = $ServicestatusTable->byUuids($serviceUuids);
        }

        $servicegroupLight = [
            'uuid'        => $servicegroup['uuid'],
            'name'        => $servicegroup['ServicegroupMapping']['display_name'],
            'description' => $servicegroup['description']
        ];

        if (empty($servicestatusByUuids)) {
            return [
                'icon'         => $this->errorIcon,
                'color'        => 'text-primary',
                'background'   => 'bg-not-monitored',
                'Servicegroup' => $servicegroupLight
            ];
        }

        $worstServiceState = array_values(
            Hash::sort($servicestatusByUuids, '{s}.current_state', 'desc')
        );

        $servicestatus = new Servicestatus($worstServiceState[0]);

        $icon = $this->serviceIcons[$servicestatus->currentState()];
        $color = $servicestatus->ServiceStatusColor();
        $background = $servicestatus->ServiceStatusBackgroundColor();


        if ($servicestatus->isAcknowledged()) {
            $icon = $this->ackIcon;
        }

        if ($servicestatus->isInDowntime()) {
            $icon = $this->downtimeIcon;
        }

        if ($servicestatus->isAcknowledged() && $servicestatus->isInDowntime()) {
            $icon = $this->ackAndDowntimeIcon;
        }

        return [
            'icon'         => $icon,
            'color'        => $color,
            'background'   => $background,
            'Servicegroup' => $servicegroupLight
        ];
    }

    /**
     * @param $dependentMapsIds
     * @param HostgroupsTable $HostgroupsTable
     * @param ServicegroupsTable $ServicegroupsTable
     * @return array
     */
    public function getAllDependentMapsElements($dependentMapsIds, HostgroupsTable $HostgroupsTable, ServicegroupsTable $ServicegroupsTable) {
        if (!is_array($dependentMapsIds)) {
            $dependentMapsIds = [$dependentMapsIds];
        }

        $allDependentMapElements = $this->find()
            ->contain([
                'Mapitems'        => function (Query $query) {
                    $query->select([
                        'Mapitems.type',
                        'Mapitems.object_id',
                        'Mapitems.map_id'
                    ])->where([
                        'Mapitems.type !=' => 'map'
                    ]);
                    return $query;
                },
                'Maplines'        => function (Query $query) {
                    $query->select([
                        'Maplines.type',
                        'Maplines.object_id',
                        'Maplines.map_id'
                    ])->where([
                        'Maplines.type !=' => 'stateless'
                    ]);
                    return $query;
                },
                'Mapgadgets'      => function (Query $query) {
                    $query->select([
                        'Mapgadgets.type',
                        'Mapgadgets.object_id',
                        'Mapgadgets.map_id'
                    ]);
                    return $query;
                },
                'Mapsummaryitems' => function (Query $query) {
                    $query->select([
                        'Mapsummaryitems.type',
                        'Mapsummaryitems.object_id',
                        'Mapsummaryitems.map_id'
                    ])->where([
                        'Mapsummaryitems.type !=' => 'map'
                    ]);
                    return $query;
                },
            ])
            ->where([
                'Maps.id IN' => $dependentMapsIds
            ])
            ->all()->toArray();

        $mapElementsByCategory = [
            'host'         => [],
            'hostgroup'    => [],
            'service'      => [],
            'servicegroup' => []
        ];
        $allDependentMapElements = Hash::filter($allDependentMapElements);

        foreach ($allDependentMapElements as $allDependentMapElementArray) {
            foreach ($allDependentMapElementArray->toArray() as $mapElementKey => $mapElementData) {
                if ($mapElementKey === 'Map') {
                    continue;
                }
                if (is_array($mapElementData)) {
                    foreach ($mapElementData as $mapElement) {
                        $mapElementsByCategory[$mapElement['type']][$mapElement['object_id']] = $mapElement['object_id'];
                    }
                }
            }

        }

        $hostUuids = $mapElementsByCategory['host'];
        if (!empty($mapElementsByCategory['hostgroup'])) {

            $members = $HostgroupsTable->find()
                ->contain(['HostgroupMembers'])
                ->where([
                    'Hostgroups.uuid IN' => $mapElementsByCategory['hostgroup']
                ])
                ->disableHydration()
                ->all();

            $members = $members->toArray();

            $memberUuids = array_unique(Hash::extract($members, '{n}.hostgroup_members.{n}.host_uuid'));
            foreach ($memberUuids as $uuid) {
                $hostIds[$uuid] = $uuid;
            }
        }

        $serviceUuids = $mapElementsByCategory['service'];
        if (!empty($mapElementsByCategory['servicegroup'])) {
            $members = $ServicegroupsTable->find()
                ->contain(['ServicegroupMembers'])
                ->where([
                    'Servicegroups.uuid IN' => $mapElementsByCategory['servicegroup']
                ])
                ->disableHydration()
                ->all();

            $members = $members->toArray();

            $memberUuids = array_unique(Hash::extract($members, '{n}.servicegroup_members.{n}.service_uuid'));
            foreach ($memberUuids as $uuid) {
                $serviceUuids[$uuid] = $uuid;
            }
        }
        return [
            'hostUuids'    => $hostUuids,
            'serviceUuids' => $serviceUuids
        ];
    }

    /**
     * @param HoststatusTable $HoststatusTable
     * @param ServicestatusTable $ServicestatusTable
     * @param array $map
     * @param array $hostsUuids
     * @param array $serviceUuids
     * @return array
     */
    public function getMapInformation(HoststatusTable $HoststatusTable, ServicestatusTable $ServicestatusTable, array $map, array $hostsUuids, array $serviceUuids) {
        $map = [
            'id'    => $map[0]['id'],
            'name'  => $map[0]['name'],
            'title' => $map[0]['title']
        ];

        if (empty($hostsUuids) && empty($serviceUuids)) {
            return [
                'icon'       => $this->errorIcon,
                'color'      => 'text-primary',
                'background' => 'bg-not-monitored',
                'Map'        => $map
            ];
        }


        $hoststatusByUuids = $HoststatusTable->byUuids($hostsUuids);

        if (empty($hoststatusByUuids)) {
            $hoststatus = new Hoststatus([]);
            $icon = $this->errorIcon;
            $color = $hoststatus->HostStatusColor();
            $background = $hoststatus->HostStatusBackgroundColor();
            $iconProperty = $icon;
        } else {
            $worstHostState = array_values(
                Hash::sort($hoststatusByUuids, '{s}.current_state', 'desc')
            );
            if (!empty($worstHostState)) {
                $hoststatus = new Hoststatus($worstHostState[0]);
            }
            $icon = $this->hostIcons[$hoststatus->currentState()];
            $color = $hoststatus->HostStatusColor();
            $background = $hoststatus->HostStatusBackgroundColor();
            $iconProperty = $icon;


            if ($hoststatus->isAcknowledged()) {
                $iconProperty = $this->ackIcon;
            }

            if ($hoststatus->isInDowntime()) {
                $iconProperty = $this->downtimeIcon;
            }

            if ($hoststatus->isAcknowledged() && $hoststatus->isInDowntime()) {
                $iconProperty = $this->ackAndDowntimeIcon;
            }
            if ($hoststatus->currentState() > 0) {
                return [
                    'icon'          => $icon,
                    'icon_property' => $iconProperty,
                    'color'         => $color,
                    'background'    => $background,
                    'Map'           => $map
                ];
            }
        }

        $servicestatus = $ServicestatusTable->byUuidsWithWhere($serviceUuids, false, ['Servicestatus.current_state IN' => [1, 2, 3]]);

        if (!empty($servicestatus)) {
            $worstServiceState = array_values(
                Hash::sort($servicestatus, '{s}.current_state', 'desc')
            );
            $servicestatus = new Servicestatus($worstServiceState[0]);
            $serviceIcon = $this->serviceIcons[$servicestatus->currentState()];

            $serviceIconProperty = $serviceIcon;
            if ($servicestatus->isAcknowledged()) {
                $serviceIconProperty = $this->ackIcon;
            }

            if ($servicestatus->isInDowntime()) {
                $serviceIconProperty = $this->downtimeIcon;
            }

            if ($servicestatus->isAcknowledged() && $servicestatus->isInDowntime()) {
                $serviceIconProperty = $this->ackAndDowntimeIcon;
            }

            return [
                'icon'           => $serviceIcon,
                'icon_property'  => $serviceIconProperty,
                'isAcknowledged' => $servicestatus->isAcknowledged(),
                'isInDowntime'   => $servicestatus->isInDowntime(),
                'color'          => $servicestatus->ServiceStatusColor(),
                'background'     => $servicestatus->ServiceStatusBackgroundColor(),
                'Map'            => $map,
            ];
        }
        return [
            'icon'           => $icon,
            'icon_property'  => $iconProperty,
            'isAcknowledged' => $hoststatus->isAcknowledged(),
            'isInDowntime'   => $hoststatus->isInDowntime(),
            'color'          => $color,
            'background'     => $background,
            'Map'            => $map
        ];
    }

    /**
     * @param ServicestatusTable $ServicestatusTable
     * @param array $hoststatus
     * @return array
     */
    public function getHostInformationForSummaryIcon(ServicestatusTable $ServicestatusTable, array $hoststatus) {
        $bitMaskHostState = 0;
        $bitMaskServiceState = 0;

        $servicestatus = $ServicestatusTable->getServicesByHostUuid($hoststatus['hostname']);

        $HostView = new Hoststatus($hoststatus);

        if ((empty($hoststatus) && empty($servicestatus))) {
            return [
                'BitMaskHostState'    => $bitMaskHostState,
                'BitMaskServiceState' => $bitMaskServiceState,
                'Host'                => $HostView->toArray(),
                'hostname'            => null
            ];
        }
        $bitMaskHostState = 1 << $HostView->currentState();

        foreach ($servicestatus as $statusDetails) {
            $ServiceView = new Servicestatus($statusDetails);
            $bitMaskServiceState |= 1 << $ServiceView->currentState();
        }
        return [
            'BitMaskHostState'    => $bitMaskHostState,
            'BitMaskServiceState' => $bitMaskServiceState,
            'Host'                => $HostView->toArray(),
            'hostname'            => $hoststatus['host_mapping']['display_name']
        ];
    }

    /**
     * @param HoststatusTable $HoststatusTable
     * @param array $servicestatus
     * @return array
     */
    public function getServiceInformationForSummaryIcon(HoststatusTable $HoststatusTable, array $servicestatus) {
        $bitMaskHostState = 0;
        $bitMaskServiceState = 0;

        $hoststatus = $HoststatusTable->byUuid($servicestatus['hostname']);

        /** @var Hoststatus $HostView */
        $HostView = new Hoststatus($hoststatus);
        /** @var Servicestatus $ServiceView */
        $ServiceView = new Servicestatus($servicestatus);

        if ((empty($hoststatus) && empty($servicestatus))) {
            return [
                'BitMaskHostState'    => $bitMaskHostState,
                'BitMaskServiceState' => $bitMaskServiceState,
                'Host'                => $HostView->toArray(),
                'Service'             => $ServiceView->toArray(),
                'hostname'            => null,
                'servicename'         => null
            ];
        }
        $bitMaskHostState = 1 << $HostView->currentState();
        $bitMaskServiceState = 1 << $ServiceView->currentState();

        return [
            'BitMaskHostState'    => $bitMaskHostState,
            'BitMaskServiceState' => $bitMaskServiceState,
            'Host'                => $HostView->toArray(),
            'Service'             => $ServiceView->toArray(),
            'hostname'            => $servicestatus['host_mapping']['display_name'],
            'servicename'         => $servicestatus['service_mapping']['display_name']
        ];
    }

    /**
     * @param HoststatusTable $HoststatusTable
     * @param ServicestatusTable $ServicestatusTable
     * @param array $hostgroup
     * @return array|void
     */
    public function getHostgroupInformationForSummaryIcon(HoststatusTable $HoststatusTable, ServicestatusTable $ServicestatusTable, array $hostgroup) {
        $hostgroupLight = [
            'uuid'        => $hostgroup['uuid'],
            'name'        => $hostgroup['HostgroupMapping']['display_name'],
            'description' => $hostgroup['description']
        ];
        $bitMaskHostState = 0;
        $bitMaskServiceState = 0;

        $hostUuids = Hash::extract($hostgroup['hostgroup_members'], '{n}.host_uuid');
        $hoststatus = $HoststatusTable->byUuids($hostUuids);
        $servicestatus = $ServicestatusTable->byHostUuids($hostUuids);

        if (empty($hoststatus) && empty($servicestatus)) {
            return [
                'BitMaskHostState'    => $bitMaskHostState,
                'BitMaskServiceState' => $bitMaskServiceState,
                'Hostgroup'           => $hostgroupLight
            ];
        }
        foreach ($hoststatus as $statusDetails) {
            $bitMaskHostState |= 1 << $statusDetails['current_state'];
        }
        foreach ($servicestatus as $statusDetails) {
            $bitMaskServiceState |= 1 << $statusDetails['current_state'];
        }
        return [
            'BitMaskHostState'    => $bitMaskHostState,
            'BitMaskServiceState' => $bitMaskServiceState,
            'Hostgroup'           => $hostgroupLight
        ];
    }

    /**
     * @param HoststatusTable $HoststatusTable
     * @param ServicestatusTable $ServicestatusTable
     * @param array $servicegroup
     * @return array
     */
    public function getServicegroupInformationForSummaryIcon(HoststatusTable $HoststatusTable, ServicestatusTable $ServicestatusTable, array $servicegroup) {
        $servicegroupLight = [
            'uuid'        => (int)$servicegroup['uuid'],
            'name'        => $servicegroup['ServicegroupMapping']['display_name'],
            'description' => $servicegroup['description']
        ];
        $bitMaskHostState = 0;
        $bitMaskServiceState = 0;

        $serviceUuids = Hash::extract($servicegroup['servicegroup_members'], '{n}.service_uuid');
        $servicestatus = $ServicestatusTable->byUuids($serviceUuids);

        $hostUuids = array_unique(Hash::extract($servicestatus, '{s}.hostname'));
        $hoststatus = $HoststatusTable->byUuids($hostUuids);

        if (empty($hoststatus) && empty($servicestatus)) {
            return [
                'BitMaskHostState'    => $bitMaskHostState,
                'BitMaskServiceState' => $bitMaskServiceState,
                'Servicegroup'        => $servicegroupLight
            ];
        }
        foreach ($hoststatus as $statusDetails) {
            $bitMaskHostState |= 1 << $statusDetails['current_state'];
        }
        foreach ($servicestatus as $statusDetails) {
            $bitMaskServiceState |= 1 << $statusDetails['current_state'];
        }
        return [
            'BitMaskHostState'    => $bitMaskHostState,
            'BitMaskServiceState' => $bitMaskServiceState,
            'Servicegroup'        => $servicegroupLight
        ];
    }

    /**
     * @param HoststatusTable $HoststatusTable
     * @param ServicestatusTable $ServicestatusTable
     * @param array $map
     * @param array $hosts
     * @param array $services
     * @return array
     */
    public function getMapInformationForSummaryIcon(HoststatusTable $HoststatusTable, ServicestatusTable $ServicestatusTable, array $map, array $hosts, array $services) {
        $bitMaskHostState = 0;
        $bitMaskServiceState = 0;

        if (empty($hosts) && empty($services)) {
            return [
                'BitMaskHostState'    => $bitMaskHostState,
                'BitMaskServiceState' => $bitMaskServiceState,
                'Map'                 => $map
            ];
        }

        $hoststatus = $HoststatusTable->byUuids($hosts);
        $servicestatus = $ServicestatusTable->byUuids($services);

        if (empty($hoststatus) && empty($servicestatus)) {
            return [
                'BitMaskHostState'    => $bitMaskHostState,
                'BitMaskServiceState' => $bitMaskServiceState,
                'Map'                 => $map
            ];
        }
        foreach ($hoststatus as $statusDetails) {
            $bitMaskHostState |= 1 << $statusDetails['current_state'];
        }
        foreach ($servicestatus as $statusDetails) {
            $bitMaskServiceState |= 1 << $statusDetails['current_state'];
        }
        return [
            'BitMaskHostState'    => $bitMaskHostState,
            'BitMaskServiceState' => $bitMaskServiceState,
            'Map'                 => $map
        ];
    }

    /**
     * @param HoststatusTable $HoststatusTable
     * @param ServicestatusTable $ServicestatusTable
     * @param array $hostgroup
     * @return array
     */
    public function getHostgroupSummary(HoststatusTable $HoststatusTable, ServicestatusTable $ServicestatusTable, array $hostgroup) {
        $hostUuids = Hash::extract($hostgroup['hostgroup_members'], '{n}.host_uuid');

        $hoststatusByUuids = $HoststatusTable->byUuids($hostUuids);
        $hostStateSummary = $HoststatusTable->getHostStateSummaryForMaps($hoststatusByUuids, false);

        $hoststatusResult = [];
        $cumulatedHostState = -1;
        $cumulatedServiceState = null;
        $allServiceStatus = [];
        $totalServiceStateSummary = [
            'state' => [
                0 => 0,
                1 => 0,
                2 => 0,
                3 => 0,
            ],
            'total' => 0
        ];

        $hostIdsGroupByState = [
            0 => [],
            1 => [],
            2 => []
        ];

        $serviceIdsGroupByState = [
            0 => [],
            1 => [],
            2 => [],
            3 => []
        ];

        foreach ($hostgroup['hostgroup_members'] as $host) {
            $host['name'] = $host['HostMapping']['display_name'];
            $hostUuid = $host['host_uuid'];

            if (isset($hoststatusByUuids[$hostUuid])) {
                $Hoststatus = new Hoststatus(
                    $hoststatusByUuids[$hostUuid]
                );
                $hostIdsGroupByState[$Hoststatus->currentState()][] = $hostUuid;

                if ($Hoststatus->currentState() > $cumulatedHostState) {
                    $cumulatedHostState = $Hoststatus->currentState();
                }
            } else {
                $Hoststatus = new Hoststatus([]);
            }

            $services = $ServicestatusTable->getServicesByHostUuid($hostUuid);

            $servicesUuids = Hash::extract($services, '{n}.service_description');
            $servicesIdsByUuid = Hash::combine($services, '{n}.service_description', '{n}.service_description');
            $servicestatusResults = $ServicestatusTable->byUuids($servicesUuids);

            $serviceIdsGroupByStatePerHost = [
                0 => [],
                1 => [],
                2 => [],
                3 => []
            ];
            foreach ($servicestatusResults as $serviceUuid => $servicestatusResult) {
                $allServiceStatus[] = $servicestatusResult['current_state'];
                $serviceIdsGroupByState[$servicestatusResult['current_state']][] = $servicesIdsByUuid[$serviceUuid];
                $serviceIdsGroupByStatePerHost[$servicestatusResult['current_state']][] = $servicesIdsByUuid[$serviceUuid];
            }

            $ServicestatusObjects = Servicestatus::fromServicestatusByUuid($servicestatusResults);
            $serviceStateSummary = ServiceStateSummary::getServiceStateSummary($ServicestatusObjects, false);

            $hoststatusResult[] = [
                'Host'                   => $host,
                'Hoststatus'             => $Hoststatus->toArray(),
                'ServiceSummary'         => $serviceStateSummary,
                'ServiceIdsGroupByState' => $serviceIdsGroupByStatePerHost
            ];

            foreach ($serviceStateSummary['state'] as $state => $stateValue) {
                $totalServiceStateSummary['state'][$state] += $stateValue;
            }
            $totalServiceStateSummary['total'] += $serviceStateSummary['total'];
        }
        $hoststatusResult = Hash::sort($hoststatusResult, '{s}.currentState', 'desc');

        $hostgroup = [
            'uuid'                => $hostgroup['uuid'],
            'name'                => $hostgroup['HostgroupMapping']['display_name'],
            'description'         => $hostgroup['description'],
            'HostSummary'         => $hostStateSummary,
            'TotalServiceSummary' => $totalServiceStateSummary
        ];

        if ($cumulatedHostState > 0) {
            $CumulatedHostStatus = new Hoststatus([
                'current_state' => $cumulatedHostState
            ]);
            $CumulatedHumanState = $CumulatedHostStatus->toArray()['humanState'];
        } else {
            if (!empty($allServiceStatus)) {
                $cumulatedServiceState = (int)max($allServiceStatus);
            }
            $CumulatedServiceStatus = new Servicestatus([
                'current_state' => $cumulatedServiceState
            ]);
            $CumulatedHumanState = $CumulatedServiceStatus->toArray()['humanState'];
        }
        return [
            'Hostgroup'              => $hostgroup,
            'Hosts'                  => $hoststatusResult,
            'CumulatedHumanState'    => $CumulatedHumanState,
            'HostIdsGroupByState'    => $hostIdsGroupByState,
            'ServiceIdsGroupByState' => $serviceIdsGroupByState
        ];
    }

    /**
     * @param \App\Model\Entity\Hoststatus $hoststatus
     * @param ServicestatusTable $ServicestatusTable
     * @param UserTime $UserTime
     * @return array
     */
    public function getHostSummary(\App\Model\Entity\Hoststatus $hoststatus, ServicestatusTable $ServicestatusTable, UserTime $UserTime) {
        $Hoststatus = new Hoststatus([], $UserTime);

        if (!empty($hoststatus)) {
            $Hoststatus = new Hoststatus($hoststatus->toArray(), $UserTime);
        }

        //Check services for cumulated state (only if host is up)
        $servicesStatus = $ServicestatusTable->getServicesByHostUuid($hoststatus->get('hostname'));

        $servicesStatusWithKey = [];
        foreach ($servicesStatus as $servicesStatusData) {
            $service =
            $servicesStatusWithKey[] = $servicesStatusData;
        }
        $ServicestatusObjects = Servicestatus::fromServicestatusByUuid($servicesStatusWithKey);
        $ServiceSummary = ServiceStateSummary::getServiceStateSummary($ServicestatusObjects, false);
        $serviceIdsGroupByState = [
            0 => [],
            1 => [],
            2 => [],
            3 => []
        ];

        foreach ($servicesStatus as $service) {
            $serviceIdsGroupByState[$service['current_state']][] = $service['service_description'];
            $ServiceStatus = new Servicestatus($service);
            $servicesResult[] = [
                'Service'       => [
                    'display_name' => $service['service_mapping']['display_name'],
                    'uuid'         => $service['service_description']
                ],
                'Servicestatus' => $ServiceStatus->toArray()
            ];

        }

        $servicesResult = Hash::sort($servicesResult, '{s}.Servicestatus.currentState', 'desc');

        return [
            'Host'                   => $hoststatus->get('host_mapping')->toArray(),
            'Hoststatus'             => $Hoststatus->toArray(),
            'Services'               => $servicesResult,
            'ServiceSummary'         => $ServiceSummary,
            'ServiceIdsGroupByState' => $serviceIdsGroupByState
        ];
    }

    /**
     * @param \App\Model\Entity\Hoststatus $hoststatus
     * @param \App\Model\Entity\Servicestatus $servicestatus
     * @param UserTime $UserTime
     * @return array
     */
    public function getServiceSummary(\App\Model\Entity\Hoststatus $hoststatus, \App\Model\Entity\Servicestatus $servicestatus, UserTime $UserTime) {
        $Hoststatus = new Hoststatus(
            $hoststatus->toArray(),
            $UserTime
        );

        if (!empty($servicestatus)) {
            $Servicestatus = new Servicestatus(
                $servicestatus->toArray(),
                $UserTime
            );
        } else {
            $Servicestatus = new Servicestatus([]);
        }

        return [
            'Host'          => $hoststatus->get('host_mapping')->toArray(),
            'Hoststatus'    => $Hoststatus->toArray(),
            'Servicestatus' => $Servicestatus->toArray(),
            'Service'       => $servicestatus->get('service_mapping')->toArray()
        ];
    }

    /**
     * @param ServicestatusTable $ServicestatusTable
     * @param array $servicegroup
     * @return array
     */
    public function getServicegroupSummary(ServicestatusTable $ServicestatusTable, array $servicegroup) {
        $serviceUuids = Hash::extract($servicegroup['servicegroup_members'], '{n}.service_uuid');

        $servicestatusResults = $ServicestatusTable->byUuids($serviceUuids);
        $ServicestatusObjects = Servicestatus::fromServicestatusByUuid($servicestatusResults);
        $serviceStateSummary = ServiceStateSummary::getServiceStateSummary($ServicestatusObjects, false);
        $serviceIdsGroupByState = [
            0 => [],
            1 => [],
            2 => [],
            3 => []
        ];
        $cumulatedServiceState = null;
        $servicesResult = [];
        foreach ($servicegroup['servicegroup_members'] as $service) {
            $serviceUuid = $service['service_uuid'];

            if (isset($servicestatusResults[$serviceUuid])) {
                $Servicestatus = new Servicestatus(
                    $servicestatusResults[$serviceUuid]
                );
                $serviceIdsGroupByState[$Servicestatus->currentState()][] = $serviceUuid;

            } else {
                $Servicestatus = new Servicestatus([]);
            }
            $servicesResult[] = [
                'Service'       => $service,
                'Servicestatus' => $Servicestatus->toArray(),
                'Host'          => $service['HostMapping']
            ];
        }
        $servicesResult = Hash::sort($servicesResult, '{s}.currentState', 'desc');
        if (!empty($servicestatusResults)) {
            $cumulatedServiceState = Hash::apply($servicestatusResults, '{s}.current_state', 'max');
        }
        $CumulatedServiceStatus = new Servicestatus([
            'current_state' => $cumulatedServiceState
        ]);

        $servicegroup = [
            'uuid'        => $servicegroup['uuid'],
            'name'        => $servicegroup['ServicegroupMapping']['display_name'],
            'description' => $servicegroup['description']
        ];

        return [
            'Servicegroup'           => $servicegroup,
            'ServiceSummary'         => $serviceStateSummary,
            'Services'               => $servicesResult,
            'CumulatedHumanState'    => $CumulatedServiceStatus->toArray()['humanState'],
            'ServiceIdsGroupByState' => $serviceIdsGroupByState
        ];
    }

    /**
     * @param HoststatusTable $HoststatusTable
     * @param ServicestatusTable $ServicestatusTable
     * @param array $map
     * @param array $hosts
     * @param array $services
     * @param UserTime $UserTime
     * @param bool $summaryStateItem
     * @return array
     */
    public function getMapSummary(HoststatusTable $HoststatusTable, ServicestatusTable $ServicestatusTable, array $map, array $hosts, array $services, UserTime $UserTime, bool $summaryStateItem) {
        $cumulatedHostState = null;
        $cumulatedServiceState = null;
        $hoststatusResult = [];
        $cumulatedHostState = -1;
        $cumulatedServiceState = null;
        $allServiceStatus = [];
        $totalServiceStateSummary = [
            'state' => [
                0 => 0,
                1 => 0,
                2 => 0,
                3 => 0,
            ],
            'total' => 0
        ];

        $hostIdsGroupByState = [
            0 => [],
            1 => [],
            2 => []
        ];

        $serviceIdsGroupByState = [
            0 => [],
            1 => [],
            2 => [],
            3 => []
        ];

        if (empty($hosts) && empty($services)) {
            return [
                'Map'                    => [
                    'name'      => $map['name'],
                    'title'     => $map['title'],
                    'object_id' => $map['id']
                ],
                'HostSummary'            => [
                    'state' => [
                        0 => 0,
                        1 => 0,
                        2 => 0
                    ],
                    'total' => 0
                ],
                'ServiceSummary'         => [
                    'state' => [
                        0 => 0,
                        1 => 0,
                        2 => 0,
                        3 => 0
                    ],
                    'total' => 0
                ],
                'CumulatedHumanState'    => __('Not in monitoring'),
                'NotOkHosts'             => 0,
                'NotOkServices'          => 0,
                'HostIdsGroupByState'    => $hostIdsGroupByState,
                'ServiceIdsGroupByState' => $serviceIdsGroupByState
            ];
        }

        $counterForNotOkHostAndService = 0;
        $limitForNotOkHostAndService = 20;


        $hostUuids = array_values($hosts);

        $hoststatusByUuids = $HoststatusTable->byUuidsWithMapping($hostUuids);
        $hostStateSummary = $HoststatusTable->getHostStateSummaryForMaps($hoststatusByUuids, false);


        $servicesUuids = array_values($services);

        $servicestatusResults = $ServicestatusTable->byUuidsWithMapping($servicesUuids);
        $ServicestatusObjects = Servicestatus::fromServicestatusByUuid($servicestatusResults);
        $serviceStateSummary = ServiceStateSummary::getServiceStateSummary($ServicestatusObjects, false);

        if (!empty($hoststatusByUuids)) {
            $worstHostState = array_values(
                $hoststatusByUuids = Hash::sort($hoststatusByUuids, '{s}.current_state', 'desc')
            );
            $cumulatedHostState = (int)$worstHostState[0]['current_state'];
            foreach ($hoststatusByUuids as $hostUuid => $hoststatusByUuid) {
                $hostStatus = new Hoststatus($hoststatusByUuid, $UserTime);
                $currentHostState = $hostStatus->currentState();

                $hostIdsGroupByState[$currentHostState][] = $hosts[$hostUuid];
                if ($counterForNotOkHostAndService <= $limitForNotOkHostAndService && $currentHostState > 0) {
                    $notOkHosts[] = [
                        'Hoststatus' => $hostStatus->toArray(),
                        'Host'       => $hoststatusByUuid
                    ];
                    $counterForNotOkHostAndService++;
                }
            }
        }

        if (!empty($servicestatusResults)) {
            $worstServiceState = array_values(
                $servicestatusResults = Hash::sort($servicestatusResults, '{s}.current_state', 'desc')
            );
            $cumulatedServiceState = (int)$worstServiceState[0]['current_state'];
            foreach ($servicestatusResults as $serviceUuid => $servicestatusByUuid) {
                $serviceStatus = new Servicestatus($servicestatusByUuid, $UserTime);
                $currentServiceState = $serviceStatus->currentState();
                $serviceIdsGroupByState[$currentServiceState][] = $services[$serviceUuid];

                if ($counterForNotOkHostAndService <= $limitForNotOkHostAndService && $currentServiceState > 0) {
                    $notOkServices[] = [
                        'Servicestatus' => $serviceStatus->toArray(),
                        'Service'       => $servicestatusByUuid
                    ];
                    $counterForNotOkHostAndService++;
                }
            }
        }

        $CumulatedHostStatus = new Hoststatus([
            'current_state' => $cumulatedHostState
        ]);

        $CumulatedHumanState = $CumulatedHostStatus->toArray()['humanState'];
        if (($cumulatedHostState === 0 || is_null($cumulatedHostState)) && !is_null($cumulatedServiceState)) {
            $CumulatedServiceStatus = new Servicestatus([
                'current_state' => $cumulatedServiceState
            ]);
            $CumulatedHumanState = $CumulatedServiceStatus->toArray()['humanState'];
        }

        $map = [
            'id'        => $map['id'],
            'name'      => $map['name'],
            'title'     => $map['title'],
            'object_id' => ($summaryStateItem) ? $map['Mapsummaryitems']['object_id'] : $map['Mapitems']['object_id']
        ];

        return [
            'Map'                    => $map,
            'HostSummary'            => $hostStateSummary,
            'ServiceSummary'         => $serviceStateSummary,
            'CumulatedHumanState'    => $CumulatedHumanState,
            'NotOkHosts'             => $notOkHosts ?? [],
            'NotOkServices'          => $notOkServices ?? [],
            'HostIdsGroupByState'    => $hostIdsGroupByState,
            'ServiceIdsGroupByState' => $serviceIdsGroupByState
        ];
    }
}
