<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace App\Command;

use App\Lib\SystemConfig;
use App\Model\Table\HostgroupMembersTable;
use App\Model\Table\HostgroupsTable;
use App\Model\Table\MappingTable;
use App\Model\Table\ServicegroupMembersTable;
use App\Model\Table\ServicegroupsTable;
use Cake\Console\Arguments;
use Cake\Command\Command;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Log\Log;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use MapModule\Model\Table\MapgadgetsTable;
use MapModule\Model\Table\MapiconsTable;
use MapModule\Model\Table\MapitemsTable;
use MapModule\Model\Table\MaplinesTable;
use MapModule\Model\Table\MapsTable;
use MapModule\Model\Table\MapsToRotationsTable;
use MapModule\Model\Table\MapsummaryitemsTable;
use MapModule\Model\Table\MaptextsTable;
use MapModule\Model\Table\RotationsTable;
use Symfony\Component\Filesystem\Filesystem;

/**
 * MappingImport command.
 */
class MappingImportCommand extends Command {

    public function initialize(): void {
        \Cake\Cache\Cache::disable();
        \Cake\Cache\Cache::clear();
        parent::initialize();
    }

    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/3.0/en/console-and-shells/commands.html#defining-arguments-and-options
     *
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);

        return $parser;
    }

    /**
     * Implement this method with your command's logic.
     *
     * @param \Cake\Console\Arguments $args The command arguments.
     * @param \Cake\Console\ConsoleIo $io The console io
     * @return null|void|int The exit code or null for success
     * @throws \Exception
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        $SystemConfig = new SystemConfig();

        $config = [
            'mapping_file'  => $SystemConfig->getMappingFile(),
            'map_json_file' => $SystemConfig->getMapJsonFile(),
            'map_zip_file'  => $SystemConfig->getMapZipFile()
        ];

        if (!isset($config['mapping_file']) || !file_exists($config['mapping_file'])) {
            throw new \RuntimeException(sprintf('Mapping file %s not found!', $config['mapping_file']));
        }

        /** @var MappingTable $MappingTable */
        $MappingTable = TableRegistry::getTableLocator()->get('Mapping');

        $sqlStatements = $MappingTable->getSchema()->truncateSql($MappingTable->getConnection());
        foreach ($sqlStatements as $sqlStatement) {
            $MappingTable->getConnection()->execute($sqlStatement)->execute();
        }

        $data = file_get_contents($config['mapping_file']);
        $data = json_decode($data, true);

        foreach ($data as $objectType => $objects) {
            if (!is_numeric($objectType)) {
                //hostgroups and servicegroups - to be done for MapModule integration on satellite - ignore for now...
                continue;
            }

            $entities = [];
            foreach ($objects as $uuid => $displayName) {
                $entity = $MappingTable->newEntity([
                    'uuid'         => $uuid,
                    'display_name' => $displayName,
                    'type_id'      => $objectType
                ]);
                $entities[] = $entity;

                if (sizeof($entities) > 500) {
                    $io->verbose('Insert ' . sizeof($entities) . ' records');
                    $MappingTable->saveMany($entities);
                    $entities = [];
                }
            }

            if (sizeof($entities) > 0) {
                $io->verbose('Insert ' . sizeof($entities) . ' records');
                $MappingTable->saveMany($entities);
                $entities = [];
            }
        }

        //Import hostgroup members
        /** @var HostgroupsTable $HostgroupsTable */
        $HostgroupsTable = TableRegistry::getTableLocator()->get('Hostgroups');
        /** @var HostgroupMembersTable $HostgroupMembersTable */
        $HostgroupMembersTable = TableRegistry::getTableLocator()->get('HostgroupMembers');

        $this->truncateTable($HostgroupsTable);
        $this->truncateTable($HostgroupMembersTable);

        if (isset($data['hostgroups']) && is_array($data['hostgroups'])) {
            foreach ($data['hostgroups'] as $hostgroup) {
                $hostgroupRecord = [
                    'uuid'              => $hostgroup['uuid'],
                    'description'       => $hostgroup['description'] ?? '',
                    'hostgroup_members' => []
                ];

                foreach ($hostgroup['members'] as $hostUuid) {
                    $hostgroupRecord['hostgroup_members'][] = [
                        'host_uuid' => $hostUuid
                    ];
                }

                $entity = $HostgroupsTable->newEmptyEntity();
                $entity = $HostgroupsTable->patchEntity($entity, $hostgroupRecord);
                $io->verbose('Insert host group ' . $hostgroup['name'] . ' (' . $hostgroup['uuid'] . ')');
                $HostgroupsTable->save($entity);
            }
        }

        //Import servicegroup members
        /** @var ServicegroupsTable $ServicegroupsTable */
        $ServicegroupsTable = TableRegistry::getTableLocator()->get('Servicegroups');
        /** @var ServicegroupMembersTable $ServicegroupMembersTable */
        $ServicegroupMembersTable = TableRegistry::getTableLocator()->get('ServicegroupMembers');

        $this->truncateTable($ServicegroupsTable);
        $this->truncateTable($ServicegroupMembersTable);

        if (isset($data['servicegroups']) && is_array($data['servicegroups'])) {
            foreach ($data['servicegroups'] as $servicegroup) {
                $servicegroupRecord = [
                    'uuid'                 => $servicegroup['uuid'],
                    'description'          => $servicegroup['description'] ?? '',
                    'servicegroup_members' => []
                ];

                foreach ($servicegroup['members'] as $serviceUuid) {
                    $servicegroupRecord['servicegroup_members'][] = [
                        'service_uuid' => $serviceUuid
                    ];
                }

                $entity = $ServicegroupsTable->newEmptyEntity();
                $entity = $ServicegroupsTable->patchEntity($entity, $servicegroupRecord);
                $io->verbose('Insert service group ' . $servicegroup['name'] . ' (' . $servicegroup['uuid'] . ')');
                $ServicegroupsTable->save($entity);
            }
        }

        if (\Cake\Core\Plugin::isLoaded('MapModule')) {
            if (file_exists($config['map_json_file'])) {
                $io->verbose('Start import of maps from maps.json file');

                $json = file_get_contents($config['map_json_file']);
                $json = json_decode($json, true);

                /** @var MapsTable $MapsTable */
                $MapsTable = TableRegistry::getTableLocator()->get('MapModule.Maps');

                /** @var MapgadgetsTable $MapgadgetsTable */
                $MapgadgetsTable = TableRegistry::getTableLocator()->get('MapModule.Mapgadgets');
                /** @var MapiconsTable $MapiconsTable */
                $MapiconsTable = TableRegistry::getTableLocator()->get('MapModule.Mapicons');
                /** @var MapitemsTable $MapitemsTable */
                $MapitemsTable = TableRegistry::getTableLocator()->get('MapModule.Mapitems');
                /** @var MaplinesTable $MaplinesTable */
                $MaplinesTable = TableRegistry::getTableLocator()->get('MapModule.Maplines');
                /** @var MapsummaryitemsTable $MapsummaryitemsTable */
                $MapsummaryitemsTable = TableRegistry::getTableLocator()->get('MapModule.Mapsummaryitems');
                /** @var MaptextsTable $MaptextsTable */
                $MaptextsTable = TableRegistry::getTableLocator()->get('MapModule.Maptexts');

                $this->truncateTable($MapsTable);
                $this->truncateTable($MapgadgetsTable);
                $this->truncateTable($MapiconsTable);
                $this->truncateTable($MapitemsTable);
                $this->truncateTable($MaplinesTable);
                $this->truncateTable($MapsummaryitemsTable);
                $this->truncateTable($MaptextsTable);

                if (isset($json['maps']) && is_array($json['maps'])) {
                    foreach ($json['maps'] as $map) {
                        $entity = $MapsTable->newEntity($map);
                        $MapsTable->save($entity);

                        if ($entity->hasErrors()) {
                            Log::error('Error on importing Map! ' . json_encode($entity->getErrors()));
                        }
                    }
                }

                //Import rotations
                /** @var RotationsTable $RotationsTable */
                $RotationsTable = TableRegistry::getTableLocator()->get('MapModule.Rotations');
                /** @var MapsToRotationsTable $MapsToRotationsTable */
                $MapsToRotationsTable = TableRegistry::getTableLocator()->get('MapModule.MapsToRotations');

                $this->truncateTable($RotationsTable);
                $this->truncateTable($MapsToRotationsTable);

                if (isset($json['rotations']) && is_array($json['rotations'])) {
                    foreach ($json['rotations'] as $rotation) {
                        $entity = $RotationsTable->newEntity($rotation);
                        $RotationsTable->save($entity);

                        if ($entity->hasErrors()) {
                            Log::error('Error on importing Rotation! ' . json_encode($entity->getErrors()));
                        }
                    }
                }


            } else {
                $io->verbose(sprintf('File "%s" does not exists', $config['map_json_file']));
            }

            if (file_exists($config['map_zip_file'])) {
                $io->verbose('Copy data from maps.zip file');
                $basePath = APP . '../' . 'plugins' . DS . 'MapModule' . DS . 'webroot' . DS . 'img' . DS;

                if (!is_dir($basePath)) {
                    mkdir($basePath);
                }

                //Delete old images
                $filesystem = new Filesystem();
                $filesystem->remove([
                    $basePath . 'backgrounds',
                    $basePath . 'icons',
                    $basePath . 'items'

                ]);

                //unzip and copy new images to webroot
                $zip = new \ZipArchive();
                if ($zip->open($config['map_zip_file']) === true) {
                    $zip->extractTo($basePath);
                    $zip->close();
                } else {
                    $io->verbose('Could not open zip file.');
                }

            } else {
                $io->verbose(sprintf('File "%s" does not exists', $config['map_zip_file']));
            }
        }

    }

    private function truncateTable(Table $Table) {
        $sqlStatements = $Table->getSchema()->truncateSql($Table->getConnection());
        foreach ($sqlStatements as $sqlStatement) {
            $Table->getConnection()->execute($sqlStatement)->execute();
        }
    }
}
