<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.
declare(strict_types=1);

namespace App\Controller;

use App\Model\Table\HoststatusTable;
use App\Model\Table\ServicestatusTable;
use Cake\Cache\Cache;
use Cake\Http\Exception\BadRequestException;
use Cake\ORM\TableRegistry;
use DateTime;
use DateTimeZone;
use Exception;
use itnovum\openITCOCKPIT\Core\Menu\Menu;
use itnovum\openITCOCKPIT\Core\ValueObjects\User;
use itnovum\openITCOCKPIT\Core\Views\HostAndServiceSummaryIcon;

/**
 * Static content controller
 *
 * This controller will render views from Template/Pages/
 *
 * @link http://book.cakephp.org/3.0/en/controllers/pages-controller.html
 */
class AngularController extends AppController {

    /****************************
     *       AJAX METHODS       *
     ****************************/

    public function menu() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template
            return;
        }

        $user = $this->getUser();
        $cacheKey = sprintf('Menu_%s', $user->get('id'));

        if (!Cache::read($cacheKey, 'permissions')) {
            $Menu = new Menu($this->PERMISSIONS);
            $jsonMenu = [];
            foreach ($Menu->getMenuItems() as $headline) {
                $jsonMenu[] = $headline->toArray();
            }

            Cache::write($cacheKey, $jsonMenu, 'permissions');
        }
        $session = $this->request->getSession();
        $session->close();

        $menu = Cache::read($cacheKey, 'permissions');

        $this->set('menu', $menu);
        $this->viewBuilder()->setOption('serialize', ['menu']);
    }

    /**
     * @throws Exception
     */
    public function user_timezone() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template
            return;
        }

        $session = $this->request->getSession();
        $session->close();

        $user = $this->getUser();
        $userTimezone = $user->get('timezone');
        $userTimezone = $userTimezone ?? 'Europe/Berlin';
        if (strlen($userTimezone) < 2) {
            $userTimezone = 'Europe/Berlin';
        }
        $UserTime = new DateTime($userTimezone);
        $ServerTime = new DateTime();
        $ServerTimeZone = new DateTimeZone($ServerTime->getTimezone()->getName());
        $timezone = [
            'user_timezone'              => $userTimezone,
            'user_time_to_server_offset' => $this->get_timezone_offset($ServerTimeZone->getName(), $userTimezone),
            'user_offset'                => $UserTime->getOffset(),
            'server_time_utc'            => time(),
            'server_time'                => date('F d, Y H:i:s'),
            'server_timezone_offset'     => $ServerTime->getOffset(),
            //ISO 8601
            'server_time_iso'            => date('c'),
            'server_timezone'            => $ServerTimeZone->getName()
        ];
        $this->set('timezone', $timezone);
        $this->viewBuilder()->setOption('serialize', ['timezone']);
    }

    /**
     * @param $remote_tz
     * @param null $origin_tz
     * @return bool|int
     * @throws Exception
     */
    private function get_timezone_offset($remote_tz, $origin_tz = null) {
        if ($origin_tz === null) {
            if (!is_string($origin_tz = date_default_timezone_get())) {
                return false; // A UTC timestamp was returned -- bail out!
            }
        }
        $origin_dtz = new DateTimeZone($origin_tz);
        $remote_dtz = new DateTimeZone($remote_tz);
        $origin_dt = new DateTime("now", $origin_dtz);
        $remote_dt = new DateTime("now", $remote_dtz);
        $offset = $origin_dtz->getOffset($origin_dt) - $remote_dtz->getOffset($remote_dt);
        return $offset;
    }

    public function menustats() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template
            return;
        }


        $user = $this->getUser();
        $showstatsinmenu = (bool)$user->get('showstatsinmenu');

        $hoststatusCount = [
            '1' => 0,
            '2' => 0,
        ];
        $servicestatusCount = [
            '1' => 0,
            '2' => 0,
            '3' => 0,
        ];

        if ($showstatsinmenu) {
            /** @var HoststatusTable $HoststatusTable */
            $HoststatusTable = TableRegistry::getTableLocator()->get('Hoststatus');
            /** @var ServicestatusTable $ServicestatusTable */
            $ServicestatusTable = TableRegistry::getTableLocator()->get('Servicestatus');

            $hoststatusCount = $HoststatusTable->getHoststatusCountStatusengine3(false);
            $servicestatusCount = $ServicestatusTable->getServicestatusCountStatusengine3(false);
        }
        $this->set('showstatsinmenu', $showstatsinmenu);
        $this->set('hoststatusCount', $hoststatusCount);
        $this->set('servicestatusCount', $servicestatusCount);
        $this->viewBuilder()->setOption('serialize', ['showstatsinmenu', 'hoststatusCount', 'servicestatusCount']);
    }

    public function topSearch() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template
            return;
        }

        if ($this->request->is('post')) {
            //Search request
            $type = $this->request->getData('type');
            $searchStr = $this->request->getData('searchStr');

            if ($type !== 'uuid') {
                throw new BadRequestException('Unknown type');
            }


            /** @var HoststatusTable $HoststatusTable */
            $HoststatusTable = TableRegistry::getTableLocator()->get('Hoststatus');
            $result = $HoststatusTable->byUuid($searchStr);

            $state = null;
            $uuid = null;
            $uuidService = null; //For ServicesBrowser
            $hasPermission = false;
            if ($result !== null) {
                $state = 'HostsBrowser';
                $uuid = $result->get('hostname');
                $hasPermission = $this->hasPermission('index', 'Hosts', '');

            }

            if ($uuid === null) {
                //No Element found yet

                /** @var ServicestatusTable $ServicestatusTable */
                $ServicestatusTable = TableRegistry::getTableLocator()->get('Servicestatus');
                $result = $ServicestatusTable->byUuid($searchStr);
                if ($result !== null) {
                    $state = 'ServicesBrowser';
                    $uuid = $result->get('hostname');
                    $uuidService = $result->get('service_description');
                    $hasPermission = $this->hasPermission('index', 'Services', '');
                }
            }


            if ($uuid !== null && $state !== null) {
                $this->set('hasPermission', $hasPermission);

                if (!$hasPermission) {
                    $this->set('message', __('You are not permitted to access this object.'));
                    $this->viewBuilder()->setOption('serialize', [
                        'hasPermission',
                        'message'
                    ]);
                    $this->response = $this->response->withStatus(403);
                    return;
                }

                $this->set('state', $state);
                $this->set('uuid', $uuid);
                $this->set('uuidService', $uuidService);
                $this->viewBuilder()->setOption('serialize', [
                    'state',
                    'uuid',
                    'uuidService',
                    'hasPermission'
                ]);
                return;
            }

        }

        $this->set('message', __('Object could not be found.'));
        $this->viewBuilder()->setOption('serialize', [
            'message'
        ]);
        $this->response = $this->response->withStatus(404);
    }

    public function downtime_host() {
        //Only ship HTML template
        $SystemConfig = new \App\Lib\SystemConfig();

        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();
        $this->set('User', $User);
        $this->set('UserTime', $UserTime);

        if ($this->isAngularJsRequest()) {
            $downtimetypeId = $SystemConfig->getPreselectedDowntimetype();
            $this->set('preselectedDowntimetype', $downtimetypeId);
            $this->viewBuilder()->setOption('serialize', ['preselectedDowntimetype']);
        }

        return;
    }

    public function getDowntimeData() {
        $SystemConfig = new \App\Lib\SystemConfig();
        $downtimetypeId = $SystemConfig->getPreselectedDowntimetype();

        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();

        $defaultValues = [
            'from_date'       => $UserTime->customFormat('d.m.Y', time()),
            'from_time'       => $UserTime->customFormat('H:i', time()),
            'to_date'         => $UserTime->customFormat('d.m.Y', time()),
            'to_time'         => $UserTime->customFormat('H:i', time() + 60 * 15),
            'duration'        => 15,
            'comment'         => __('In maintenance'),
            'downtimetype_id' => $downtimetypeId
        ];

        $this->set('defaultValues', $defaultValues);
        $this->viewBuilder()->setOption('serialize', ['defaultValues']);
    }

    /********************************
     *       TEMPLATE METHODS       *
     ********************************/

    public function paginator() {
        //Only ship HTML template
    }

    public function scroll() {
        //Only ship HTML template
    }

    public function mass_delete() {
        //Only ship HTML template
    }

    public function sidebar() {
        //Only ship HTML template
        return;
    }

    public function not_found() {
        //Only ship HTML template
        return;
    }

    public function forbidden() {
        //Only ship HTML template
        return;
    }

    public function reload_required() {
        //Only ship HTML template
        return;
    }

    public function reschedule_host() {
        //Only ship HTML template
        return;
    }

    public function acknowledge_host() {
        //Only ship HTML template
        return;
    }

    public function submit_host_result() {
        //Only ship HTML template
        return;
    }

    public function disable_host_flap_detection() {
        //Only ship HTML template
        return;
    }

    public function enable_host_flap_detection() {
        //Only ship HTML template
        return;
    }

    public function executing() {
        //Only ship HTML template
        $id = $this->request->getQuery('id', 'angularExecutingModal');

        $this->set('id', $id);
    }

    public function reschedule_service() {
        //Only ship HTML template
        return;
    }

    public function hosticon() {
        //Only ship HTML template
        return;
    }

    public function serviceicon() {
        //Only ship HTML template
        return;
    }


    public function downtime_service() {
        //Only ship HTML template

        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();
        $this->set('User', $User);
        $this->set('UserTime', $UserTime);
        return;
    }

    public function acknowledge_service() {
        //Only ship HTML template
        return;
    }

    public function submit_service_result() {
        //Only ship HTML template
        return;
    }

    public function disable_service_flap_detection() {
        //Only ship HTML template
        return;
    }

    public function enable_service_flap_detection() {
        //Only ship HTML template
        return;
    }

    public function mass_delete_host_downtimes() {
        //Only ship HTML template
        return;
    }

    public function mass_delete_service_downtimes() {
        //Only ship HTML template
        return;
    }

    public function popover_graph() {
        //Only ship HTML template
        return;
    }
    /**
     * @param int $size
     * @param int $bitMaskHostState
     * @param int $bitMaskServiceState
     * @throws Exception
     */

    public function getHostAndServiceStateSummaryIcon($size = 100, $bitMaskHostState = 0, $bitMaskServiceState = 0) {
        $session = $this->request->getSession();
        $session->close();
        $HostAndServiceSummaryIcon = new HostAndServiceSummaryIcon($size);
        $HostAndServiceSummaryIcon->createSummaryIcon($bitMaskHostState, $bitMaskServiceState);
        $image = $HostAndServiceSummaryIcon->getImage();

        $this->disableAutoRender();
        $this->response = $this->response->withHeader('Content-Type', 'image/png');
        $stream = new \Laminas\Diactoros\CallbackStream(function () use ($image) {
            imagepng($image, null, 0);
            imagedestroy($image);
        });
        $this->response = $this->response->withBody($stream);
    }
}
