<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace itnovum\openITCOCKPIT\Core\Security;


use BadRequestException;
use Controller;
use SecurityException;
use SessionComponent;

class CSRF {

    /**
     * @var SessionComponent
     */
    protected $Session;

    /**
     * strtotime compatible string
     * Set time how long a token is valide
     * @var string
     */
    protected $csrfExpires = '+30 minutes';

    /**
     * Maximum number of tokens stored in session
     * @var int
     */
    protected $csrfLimit = 150;

    /**
     * Remove CSRF tokens after use
     * Can cause issues for Single Page applications
     * @var boolean
     */
    protected $csrfUseOnce = true;

    /**
     * Token length
     * @var int
     */
    protected $length = 16;

    /**
     * Token storage from and for $_SESSION
     * @var array
     */
    protected $tokens = [];

    /**
     * CSRF constructor.
     * @param SessionComponent $Session
     */
    public function __construct(SessionComponent $Session) {
        $this->Session = $Session;

        $tokens = $this->readTokens();
        $this->tokens = $this->expireTokens($tokens);
    }

    /**
     * @return array
     */
    protected function readTokens() {
        $csrf = $this->Session->read('_csrf');

        $tokens = [];
        if (isset($csrf['csrfTokens'])) {
            $tokens = $csrf['csrfTokens'];
            if (!is_array($tokens)) {
                $tokens = [];
            }
        }

        return $tokens;
    }

    /**
     * @param array $tokens
     * @return array
     */
    protected function expireTokens($tokens) {
        $now = time();
        foreach ($tokens as $nonce => $expires) {
            if ($expires < $now) {
                unset($tokens[$nonce]);
            }
        }
        $overflow = count($tokens) - $this->csrfLimit;
        if ($overflow > 0) {
            //Remove oldest tokens
            asort($tokens);
            $tokens = array_slice($tokens, $overflow + 1, null, true);
        }
        return $tokens;
    }

    /**
     * @param Controller $Controller
     * @return bool
     */
    public function validateCsrfToken(Controller $Controller) {
        $requestToken = $Controller->request->data('_csrfToken');

        if (!$requestToken) {
            throw new SecurityException('Missing CSRF token');
        }

        if (!isset($this->tokens[$requestToken])) {
            throw new SecurityException('CSRF token mismatch');
        }

        if ($this->tokens[$requestToken] < time()) {
            throw new SecurityException('CSRF token expired');
        }

        if (isset($this->tokens[$requestToken])) {

            if ($this->csrfUseOnce) {
                unset($this->tokens[$requestToken]);
            }
            return true;
        }

        $this->blackhole();
    }

    public function blackhole() {
        throw new BadRequestException(__d('cake_dev', 'The request has been black-holed'));
    }

    /**
     * Generate a new Token stores it to $_SESSION and return the current token
     * @return string
     */
    public function generateToken() {
        $newToken = $this->_generateToken();
        $this->tokens[$newToken] = strtotime($this->csrfExpires);

        $csrf = [
            'latestToken' => $newToken,
            'csrfTokens'  => $this->tokens
        ];

        $this->Session->write('_csrf', $csrf);

        return $newToken;
    }

    /**
     * @return string
     */
    private function _generateToken() {
        if ($this->length < 16) {
            throw new SecurityException('Token length needs to be >= 16!');
        }

        return hash('sha512', openssl_random_pseudo_bytes($this->length));
    }

}
