<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

/**
 * Class CpuTotalPercentage
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class CpuTotalPercentage extends BasePlugin implements PluginInterface {

    public function execute() {
        if ($this->checkError('cpu', 'cpu_total_percentage')) {
            return;
        }

        $requiredKeys = [
            'agent' => [
                'goos'
            ],
            'cpu'   => [
                'cpu_total_percentage',
                'cpu_total_percentage_detailed',
                'cpu_percentage'
            ]
        ];

        foreach ($requiredKeys as $mainKey => $subKeys) {
            foreach ($subKeys as $subKey) {
                if (!isset($this->json[$mainKey][$subKey])) {
                    $this->setOutput(sprintf('Key "%s.%s" not found in Agent output', $mainKey, $subKey));
                    $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
                    return;
                }
            }
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $warning = 90;
        $critical = 95;
        $perCore = 0;
        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $warning = $arg;
                    break;
                case 1:
                    $critical = $arg;
                    break;
                case 2:
                    $perCore = (bool)$arg;
                    break;
            }
        }

        $goos = $this->json['agent']['goos'];
        $cpuTotalPercent = round($this->json['cpu']['cpu_total_percentage'], 2);

        // Field overview (all values are percentage)
        // Windows: User, System, Idle
        // Linux: User, Nice, System, Iowait
        // macOS: User, Nice, System, Idle

        $fields = [
            'windows' => [
                'User', 'System', 'Idle'
            ],
            'linux'   => [
                'User', 'Nice', 'System', 'Iowait'
            ],
            'darwin'  => [
                'User', 'Nice', 'System', 'Idle'
            ]
        ];

        if (!isset($fields[$goos])) {
            $this->setOutput(sprintf('Agent is running on an unsupported operating system: %s', $goos));
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }


        $this->addPerfdata(new PerfdataGauge(
            'Total',
            $cpuTotalPercent,
            '%',
            $warning,
            $critical,
            0,
            100
        ));

        foreach ($fields[$goos] as $gaugeName) {
            if (isset($this->json['cpu']['cpu_total_percentage_detailed'][$gaugeName])) {
                $this->addPerfdata(new PerfdataGauge(
                    $gaugeName,
                    round($this->json['cpu']['cpu_total_percentage_detailed'][$gaugeName], 2),
                    '%',
                    null,
                    null,
                    0,
                    100
                ));
            }
        }

        if ($perCore) {
            foreach ($this->json['cpu']['cpu_percentage'] as $coreNumber => $cpuUsage) {
                $this->addPerfdata(new PerfdataGauge(
                    sprintf('CPU%s', $coreNumber),
                    round($cpuUsage, 2),
                    '%',
                    $warning,
                    $critical,
                    0,
                    100
                ));
            }
        }

        $output = sprintf('Total CPU usage is: %s%%', $cpuTotalPercent);
        foreach ($fields[$goos] as $gaugeName) {
            if (isset($this->json['cpu']['cpu_total_percentage_detailed'][$gaugeName])) {
                $output = sprintf(
                    '%s, %s: %s%%',
                    $output,
                    $gaugeName,
                    round($this->json['cpu']['cpu_total_percentage_detailed'][$gaugeName], 2)
                );
            }
        }

        $this->setOutput($output);


        if ($this->getRangeStatus($critical, $cpuTotalPercent)) {
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return;
        }

        if ($this->getRangeStatus($warning, $cpuTotalPercent)) {
            $this->setStatuscode(NagiosService::STATUS_WARNING);
            return;
        }

        $this->setStatuscode(NagiosService::STATUS_OK);
    }
}
