<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\Converter\Byte;
use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

/**
 * Class DiskIO
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class DiskIO extends BasePlugin implements PluginInterface {

    public function execute() {
        if ($this->checkError('disk_io')) {
            return;
        }

        $requiredKeys = [
            'disk_io' => [
                'Timestamp',           // unitx timestamp
                'Device',              // C: or sda
                'ReadIopsPerSecond',   // uint64  // Number of read iops per second
                'WriteIopsPerSecond',  // uint64  // Number of write iops per second
                'TotalIopsPerSecond',  // uint64  // Number of read and write iops per second
                'ReadBytesPerSecond',  // uint64  // Number of bytes read from disk per second
                'WriteBytesPerSecond', // uint64  // Number of bytes written to disk per second
                'TotalAvgWait',        // float64 // Total io_wait in milliseconds
                'ReadAvgWait',         // float64 // Average io_wait for read iops in milliseconds
                'WriteAvgWait',        // float64 // Average io_wait for write iops in milliseconds
                'ReadAvgSize',         // float64 // Average request size of reads in bytes
                'WriteAvgSize',        // float64 // Average request size of writes in bytes
                'LoadPercent',         // float64 // Disk load as percentage
            ]
        ];

        foreach ($requiredKeys as $mainKey => $subKeys) {
            if (isset($this->json[$mainKey])) {
                foreach ($this->json[$mainKey] as $device => $data) {
                    foreach ($subKeys as $subKey) {
                        if (!isset($this->json[$mainKey][$device][$subKey])) {
                            $this->setOutput(sprintf('Key "%s.%s.%s" not found in Agent output', $mainKey, $device, $subKey));
                            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
                            return;
                        }
                    }
                }
            }
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $warning = $arg;
                    break;
                case 1:
                    $critical = $arg;
                    break;
                case 2:
                    $device = $arg;
                    break;
            }
        }

        if (!isset($this->json['disk_io'][$device])) {
            $this->setOutput(sprintf('Device "%s" not found in Agent output', $device));
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $io = $this->json['disk_io'][$device];

        $this->setOutput(sprintf(
            'Disk load: %s%%. Read IOPS: %s, Write IOPS: %s. Write request size: %s, Total Wait: %sms',
            round($io['LoadPercent'], 2),
            round($io['ReadIopsPerSecond'], 2),
            round($io['WriteIopsPerSecond'], 2),
            Byte::asHuman($io['WriteAvgSize']),
            round($io['TotalAvgWait'], 4)
        ));

        $this->addPerfdata(new PerfdataGauge(
            'Total_Disk_Load',
            round($io['LoadPercent'], 2),
            '%',
            $warning,
            $critical,
            0,
            100
        ));

        $jsonToGauge = [
            'ReadIopsPerSecond'  => 'read_iops',
            'WriteIopsPerSecond' => 'write_iops',
            'TotalIopsPerSecond' => 'total_iops'
        ];
        foreach ($jsonToGauge as $jsonKey => $gaugeName) {
            $this->addPerfdata(new PerfdataGauge(
                $gaugeName,
                $io[$jsonKey],
                'iops'
            ));
        }

        $jsonToGauge = [
            'ReadAvgWait'  => 'read_avg_wait',
            'WriteAvgWait' => 'write_avg_wait',
            'TotalAvgWait' => 'total_avg_wait'
        ];
        foreach ($jsonToGauge as $jsonKey => $gaugeName) {
            $this->addPerfdata(new PerfdataGauge(
                $gaugeName,
                $io[$jsonKey],
                'ms'
            ));
        }

        $jsonToGauge = [
            'ReadAvgSize'  => 'read_avg_size',
            'WriteAvgSize' => 'write_avg_size',
        ];
        foreach ($jsonToGauge as $jsonKey => $gaugeName) {
            $this->addPerfdata(new PerfdataGauge(
                $gaugeName,
                $io[$jsonKey],
                'B'
            ));
        }

        $jsonToGauge = [
            'ReadBytesPerSecond'  => 'read_bytes_sec',
            'WriteBytesPerSecond' => 'write_bytes_sec',
        ];
        foreach ($jsonToGauge as $jsonKey => $gaugeName) {
            $this->addPerfdata(new PerfdataGauge(
                $gaugeName,
                $io[$jsonKey],
                'B/s'
            ));
        }

        if ($this->getRangeStatus($critical, $io['LoadPercent'])) {
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return;
        }

        if ($this->getRangeStatus($warning, $io['LoadPercent'])) {
            $this->setStatuscode(NagiosService::STATUS_WARNING);
            return;
        }


        $this->setStatuscode(NagiosService::STATUS_OK);
    }
}
