<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\Converter\Byte;
use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

/**
 * Class DiskUsage
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class DiskUsage extends BasePlugin implements PluginInterface {

    public function execute() {
        if ($this->checkError('disks')) {
            return;
        }

        $requiredKeys = [
            'disks'
        ];

        foreach ($requiredKeys as $requiredKey) {
            if (!isset($this->json[$requiredKey])) {
                $this->setOutput(sprintf('Key "%s" not found in Agent output', $requiredKey));
                $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
                return;
            }
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $warning = $arg;
                    break;
                case 1:
                    $critical = $arg;
                    break;
                case 2:
                    $mountpoint = $arg;
                    break;
                case 3:
                    $unit = $arg;
                    break;
            }
        }

        if (!in_array($unit, ['GB', '%'])) {
            $this->setOutput('Plugin arguments "unit" not found or invalid. (available options: GB, %)');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        // Remove trailing backslash for Windows Disks C:\ => C:
        // https://github.com/it-novum/openITCOCKPIT/issues/1062
        // Auto 1.x to 3.x fix.
        if (substr($mountpoint, -2) == ':\\' && strlen($mountpoint) >= 3) {
            $mountpoint = substr($mountpoint, 0, strlen($mountpoint) - 1); // Remove trailing backslash
        }

        $disk = null;
        foreach ($this->json['disks'] as $ldisk) {
            if ($ldisk['disk']['mountpoint'] === $mountpoint && !empty($ldisk['usage'])) {
                $disk = $ldisk;
                break;
            }
        }

        if (empty($disk)) {
            $this->setOutput(sprintf('No disk with mountpoint "%s" in Agent output', $mountpoint));
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return;
        }

        $diskFreePercentage = 100.0 - $disk['usage']['percent'];

        $this->addPerfdata(new PerfdataGauge(
            'Disk_Used',
            round($disk['usage']['percent'], 2),
            '%',
            ($unit === '%') ? $warning : null,
            ($unit === '%') ? $critical : null,
            0,
            100
        ));

        foreach (['used', 'total', 'free'] as $gauge) {
            $this->addPerfdata(new PerfdataGauge(
                $gauge,
                $disk['usage'][$gauge],
                'B'
            ));
        }

        $this->setOutput(sprintf(
            'Disk space used: %s/%s (%s%%). Disk space free: %s (%s%%)',
            Byte::asHuman($disk['usage']['used']),
            Byte::asHuman($disk['usage']['total']),
            round($disk['usage']['percent'], 2),
            Byte::asHuman($disk['usage']['free']),
            round($diskFreePercentage, 2)
        ));

        if ($unit === '%') {
            // Warning and critical are set as percentage (%)
            if ($this->getRangeStatus($critical, $disk['usage']['percent'])) {
                $this->setStatuscode(NagiosService::STATUS_CRITICAL);
                return;
            }

            if ($this->getRangeStatus($warning, $disk['usage']['percent'])) {
                $this->setStatuscode(NagiosService::STATUS_WARNING);
                return;
            }

            $this->setStatuscode(NagiosService::STATUS_OK);
            return;
        }

        // Warning and critical are set in GB
        $diskUsageGB = $this->byteToGb($disk['usage']['used']);
        if ($this->getRangeStatus($critical, $diskUsageGB)) {
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return;
        }

        if ($this->getRangeStatus($warning, $diskUsageGB)) {
            $this->setStatuscode(NagiosService::STATUS_WARNING);
            return;
        }

        $this->setStatuscode(NagiosService::STATUS_OK);
    }

    private function byteToGb($byte) {
        return ($byte / 1024) / 1024 / 1024;
    }
}
