<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\Converter\Byte;
use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

/**
 * Class NetIO
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class NetIO extends BasePlugin implements PluginInterface {

    public function execute() {
        if ($this->checkError('net_io')) {
            return;
        }

        $requiredKeys = [
            'net_io' => [
                'name',                 // string // Name of the network interface
                'timestamp',            // int64  // Timestamp of the last check evaluation
                'bytes_sent',           // uint64 // Number of bytes sent
                'bytes_recv',           // uint64 // Number of bytes received
                'packets_sent',         // uint64 // Number of packets sent
                'packets_recv',         // uint64 // Number of bytes received
                'errin',                // uint64 // Total number of errors while receiving
                'errout',               // uint64 // Total number of errors while sending
                'dropin',               // uint64 // Total number of incoming packets which were dropped
                'dropout',              // uint64 // Total number of outgoing packets which were dropped (always 0 on macOS and BSD)
                'avg_bytes_sent_ps',    // uint64 // Average bytes sent per second
                'avg_bytes_recv_ps',    // uint64 // Average bytes received per second
                'avg_packets_sent_ps',  // uint64 // Average packets sent per second
                'avg_packets_recv_ps',  // uint64 // Average packets received per second
                'avg_errin',            // uint64 // Average errors while receiving per second
                'avg_errout',           // uint64 // Average errors while sending per second
                'avg_dropin',           // uint64 // Average incoming dropped packets per second
                'avg_dropout',          // uint64 // Average outgoing dropped packets per second
            ]
        ];

        foreach ($requiredKeys as $mainKey => $subKeys) {
            if (isset($this->json[$mainKey])) {
                foreach ($this->json[$mainKey] as $nic => $data) {
                    foreach ($subKeys as $subKey) {
                        if (!isset($this->json[$mainKey][$nic][$subKey])) {
                            $this->setOutput(sprintf('Key "%s.%s.%s" not found in Agent output', $mainKey, $nic, $subKey));
                            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
                            return;
                        }
                    }
                }
            }
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $bytesWarning = $arg;
                    break;
                case 1:
                    $bytesCritical = $arg;
                    break;
                case 2:
                    $errorWarning = $arg;
                    break;
                case 3:
                    $errorCritical = $arg;
                    break;
                case 4:
                    $dropWarning = $arg;
                    break;
                case 5:
                    $dropCritical = $arg;
                    break;
                case 6:
                    $networkDevice = $arg;
                    break;
            }
        }

        if (!isset($this->json['net_io'][$networkDevice])) {
            $this->setOutput(sprintf('Network device "%s" not found in Agent output', $networkDevice));
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $io = $this->json['net_io'][$networkDevice];
        $tx = Byte::scaleBytesPerSecondToBitsPerSecondForHumans($io['avg_bytes_sent_ps']);
        $rx = Byte::scaleBytesPerSecondToBitsPerSecondForHumans($io['avg_bytes_recv_ps']);

        foreach (['avg_bytes_sent_ps', 'avg_bytes_recv_ps'] as $gaugName) {
            $this->addPerfdata(new PerfdataGauge(
                $gaugName,
                $io[$gaugName],
                'B/s', // bytes per second
                $bytesWarning,
                $bytesCritical
            ));
        }

        $this->addPerfdata(new PerfdataGauge(
            'avg_bytes_sent_kbits',
            ($io['avg_bytes_sent_ps'] * 8 / 1024),
            'kbits', // kbits
            ($bytesWarning * 8 / 1024),
            ($bytesCritical * 8 / 1024),
            0
        ));
        $this->addPerfdata(new PerfdataGauge(
            'avg_bytes_recv_kbits',
            ($io['avg_bytes_recv_ps'] * 8 / 1024),
            'kbits', // kbits
            ($bytesWarning * 8 / 1024),
            ($bytesCritical * 8 / 1024),
            0
        ));

        foreach (['avg_packets_sent_ps', 'avg_packets_recv_ps'] as $gaugName) {
            $this->addPerfdata(new PerfdataGauge(
                $gaugName,
                $io[$gaugName],
                ''
            ));
        }

        foreach (['avg_errin', 'avg_errout'] as $gaugName) {
            $this->addPerfdata(new PerfdataGauge(
                $gaugName,
                $io[$gaugName],
                '',
                $errorWarning,
                $errorCritical,
                0
            ));
        }

        foreach (['avg_dropin', 'avg_dropout'] as $gaugName) {
            $this->addPerfdata(new PerfdataGauge(
                $gaugName,
                $io[$gaugName],
                '',
                $dropWarning,
                $dropCritical,
                0
            ));
        }

        $output = sprintf(
            '[%s] Bandwidth: TX: %s / %s RX: %s / %s. Errors in/out per second: %s/%s Drop in/out per second: %s/%s',
            $io['name'],
            Byte::scaleBytesPerSecondForHumans($io['avg_bytes_sent_ps']),
            $tx['value'] . $tx['unit'],
            Byte::scaleBytesPerSecondForHumans($io['avg_bytes_recv_ps']),
            $rx['value'] . $rx['unit'],
            $io['avg_errin'],
            $io['avg_errout'],
            $io['avg_dropin'],
            $io['avg_dropout']
        );


        $total_bytes = $io['avg_bytes_sent_ps'] + $io['avg_bytes_recv_ps'];
        $total_error = $io['avg_errin'] + $io['avg_errout'];
        $total_drop = $io['avg_dropin'] + $io['avg_dropout'];

        $preformat = '';
        $this->setStatuscode(NagiosService::STATUS_OK);
        if ($this->getRangeStatus($bytesWarning, $total_bytes)) {
            $preformat = 'Total average bytes warning! ';
            $this->setStatuscode(NagiosService::STATUS_WARNING);
        }

        if ($this->getRangeStatus($errorWarning, $total_error)) {
            $preformat = 'Total average error warning! ';
            $this->setStatuscode(NagiosService::STATUS_WARNING);
        }

        if ($this->getRangeStatus($dropWarning, $total_drop)) {
            $preformat = 'Total average drop warning! ';
            $this->setStatuscode(NagiosService::STATUS_WARNING);
        }

        if ($this->getRangeStatus($bytesCritical, $total_bytes)) {
            $preformat = 'Total average bytes critical! ';
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
        }

        if ($this->getRangeStatus($errorCritical, $total_error)) {
            $preformat = 'Total average error critical! ';
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
        }

        if ($this->getRangeStatus($dropCritical, $total_drop)) {
            $preformat = 'Total average drop critical! ';
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
        }

        $this->setOutput($preformat . $output);
    }
}
